#!/usr/bin/perl -w
use diagnostics;
use strict;
use HTML::Entities();

use ezApp::DB;

package ezApp::Interpreter;

#---------------------------------------------------------------
#
#			Interpreter for ezApp base tag set.
#
# 	Copyright (c) 2001 Richard K. Moore.
#				richard@cyberjournal.org
#
#	This script is free software; you can redistribute it and/or
#	modify it under the same terms as Perl itself.
#
#-------------------------------------------------------------------
#
#		ezApp::Interpreter provides the base methods which support
#		ezApp's extended HTML tag set.  Additional tags can
#		be implemented either as Subscreens, or as methods
#		in an Interpreter subclass.  Subscreens override
#		all methods, and subclass methods override the base methods
#		defined here.
#
#---------------------------------------------------------------
	


	# Hashes to encode tagnames as integers, if that becomes advisable.
	# Meanwhile, they serve as hashes to test what kind of tag we've got.
	
		my @ezBlockTags = qw
			( if elsif else while repeat foreach subscreen screen );
			
		my $tagKeyVal = 5;
		my $ezBlockTagCodes = 
			{ map {$_ => $tagKeyVal++} @ezBlockTags };
			
	
		my @ezNonBlockTags = qw
		( 
			editvar persistent nonpersistent hiddenfields 
			ezbutton gotoscreen
			dbopen dbclose dbgetrecords dbupdate dbdo dblastinsertid
		);
	
		my $ezNonBlockTagCodes = 
			{ map {$_ => $tagKeyVal++} @ezNonBlockTags };
			
	
		my @HtmlTags = qw
		(
			a abbr acronym address applet area b basefont base bdo bgsound
			big blink blockquote body br button caption center cite code col
			colgroup dd del dir div dl dt em embed fieldset font form
			frame frameset h1 h2 h3 h4 h5 h6 head hr html i iframe img
			input ins isindex kbd keygen label layer legend li link map
			marquee meta menu multicol nobr nembed noframes nolayer
			noscript object ol optgroup option p param pre q s samp
			script select server small spacer span strike strong
			sub sup table tbody td textarea tfoot th thead title tr
			tt u ul var wbr
		);
	
		my $HtmlTagCodes = 
			{ map {$_ => $tagKeyVal++} @HtmlTags };
			
		my $TagCodes = {%$ezBlockTagCodes, %$ezNonBlockTagCodes, %$HtmlTagCodes};

		
		
	my $VarNamePattern = '[a-zA-Z^][a-zA-Z0-9^_]*';
	my $ParNamePattern =	'([a-z^]|[a-z^][a-z0-9^_]*[a-z0-9])';

	my $TxtPrefix = "---------------------------- output -\t\t";

#---------------------------------------------------------------
sub new 	
{
	my ($class, $job) 				= @_;
	my $self 									= {};
	bless $self, $class;
	$self->{job} 							= $job;
$self}
#----------------------------------------------------------------------
sub InterpretFile
{
	my ($self, $FileName) = @_;
	$self->{FileName}			= $FileName;
	my $job								= $self->{job};
	$self->{OutputBuffer}	= '';
	$self->{SubscreenObj}			= $self;
	$self->{Subscreens}				= {};
	$self->{RedirectFile}			= '';
	
	my ($ScreenContents, $ErrMsg);
	$job->ReadEntireFile("ezscrs/$FileName", \$ScreenContents, \$ErrMsg);
	! $ErrMsg
		|| return $job->dieClean("$ErrMsg");
		
	$self->{Blocks} = { split "\n".chr(3)."\n", $ScreenContents };
	
	# $job->trace('keys blocks='. join '|', '', keys %{$self->{Blocks}}, '' );
	
	$self->InterpretBlock(1);
}
#----------------------------------------------------------------------
sub InterpretBlock	
{
	my ($self, $BlockKey) = @_;
	my $job 							= $self->{job};
	
	! ( $job->{Died} || $self->{RedirectFile} )
		|| return;
	
	# $job->trace("\nBEGIN BLOCK - $BlockKey");
	
	my $Block = $self->{Blocks}->{$BlockKey}
		||  return $job->dieClean("Cannot find Blocks->{$BlockKey}");
		
	my @Items = split chr(2)."\n", $Block;
	
	foreach (@Items)
	{ 
		$self->InterpretItem($_);
		return
			if ( $job->{Died} || $self->{Terminator} || $self->{RedirectFile} );
	};
}
#----------------------------------------------------------------------
sub InterpretItem		# Interpret an item and return the terminating condition.
{
	my ($self, $Item) = @_;
	my $job 					= $self->{job};
	my $SubScreens		= $self->{Subscreens};
	my ($LineNumber, $Kind, $TagOrText, @params) = split chr(1), $Item;
	
	! $job->{Died}
		|| return;
	
	# $job->tracedump('InterpretItem - $LineNumber, $Kind, $TagOrText, @params',$LineNumber, $Kind, $TagOrText, @params )
	#	if $job->{fhTrace};
	
	if ($LineNumber)
	{
		$self->{LineNumber} = $LineNumber;
		# $job->trace("\nBegin Line# $LineNumber");
	};
		
	$self->{Tag} = $TagOrText;

	if ($Kind eq '<')
	{
		# A start tag.
		
		# Pick up sub-block key if there is one.
		
			( $self->{NestedBlockKey} = shift @params )
		
				if $ezBlockTagCodes->{$TagOrText}
					|| $SubScreens->{$TagOrText};
				
		# Move rest of params to $self->{Attr, AttrSeq}.
		
			$self->BuildAttr( $TagOrText, [@params] );
		
		if ( $HtmlTagCodes->{$TagOrText} )
		{
			# HTML tag... evaluate attributes and pass it on.
			
				$self->htmlTag($TagOrText);
		}
		elsif ( $SubScreens->{$TagOrText} )
		{
			$self->CallSubscreen($TagOrText);
		}
		elsif( $self->can($TagOrText) )
		{
			# ezApp tag.
			
				$self->ezAppTag($TagOrText);
				! $job->{Died}
					|| return;
		}
		else
			{ return $job->dieClean("Unrecgnized tag &lt;$TagOrText&gt;") };
	}
	elsif ($Kind eq '/')
	{
		$self->{OutputBuffer} .= "</$TagOrText>";
	}
	elsif ($Kind eq '"')
	{
		$self->TextItem($TagOrText);
	}
	else
		{ return $job->dieClean("Unknown Item kind = '$Kind'") };
}
#----------------------------------------------------------------------
sub BuildAttr			# Encode param list into $self->{Attr,AttrSeq}.
{
	my ($self, $Tag, $Params) = @_;
	my $job										= $self->{job};
	
	my ($AttrName, $AttrVal);
	my $Attr = {};
	my $AttrSeq = [];
	
	# $job->trace("\nBuildAttr");
	
	while ($AttrName = shift @$Params)
	{
		push @$AttrSeq, $AttrName;
		$AttrVal = shift @$Params;

		if ( $AttrVal =~ /^{(.+?)}$/s && $Tag !~ /^(while|foreach)$/s )
		{
			# AttrVal is an expression - evaluate it.
			
				$AttrVal = $self->ContextEval($1);
				! $job->{Died}
					|| return;
		};
		
		$Attr->{$AttrName} = $AttrVal;
		
		# $job->tracedump("\tAttrName='$AttrName', AttrVal=", $AttrVal)
		#	if $job->{fhTrace};
	};
	
	$self->{Attr} 		= $Attr;
	$self->{AttrSeq} 	= $AttrSeq;
}
#----------------------------------------------------------------------
sub htmlTag
{
	my ($self, $Tag) 			= @_;
	my $job 							= $self->{job};
	my ($Attr, $AttrSeq)	= @$self{'Attr','AttrSeq'};
	
	# $job->trace("html <$Tag>");
	
	$self->{OutputBuffer} .= "<$Tag";
	
	foreach my $AttrName (@$AttrSeq)
	{
		$self->{OutputBuffer} .= " $AttrName=";
		my $AttrVal = $Attr->{$AttrName};
		
		$self->{OutputBuffer} .= $self->QuoteAttribute($AttrVal);
		! $job->{Died}
			|| return;
	};
	
	$self->{OutputBuffer} .= '>';
}
#----------------------------------------------------------------------
sub ezAppTag		# Interpret an ezApp tag and return the termination condition.
{
	my ($self, $Tag) 	= @_;
	my $job 					= $self->{job};
	
	# $job->trace("ezApp <$Tag>");
	
	$self->$Tag();
}
#----------------------------------------------------------------------
sub CallSubscreen		# Interpret subscreen instance into current output stream.
{
	my ($self, $SubscreenName) 	= @_;
	my $job											= $self->{job};
	my $Subscreens							= $self->{Subscreens};
	my $Attr										= $self->{Attr};
			
	push @{ $self->{SubscreenStack} }, [ @$self{ qw(FileName LineNumber SubscreenObj) } ];
	
	my ($NestedBlockKey, $ParNames, $FileName, $LineNumber)
		= @{ $Subscreens->{$SubscreenName} }
			{ qw(BodyKey Pars FileName LineNumber) };
			
			
#			die 'CallSubScreen - $SubscreenName, $NestedBlockKey, $ParNames, $FileName, $LineNumber = '
#				.$job->dump($SubscreenName, $NestedBlockKey, $ParNames, $FileName, $LineNumber);

	
	
	# Instantiate a package & object for this sub call instance, and initialize its par space.
	
		my $SubscreenObj = $self->SetupParSpace($ParNames, $SubscreenName, $Attr);
		! $job->{Died}
			|| return;
			
	@$self{'FileName','LineNumber','SubscreenObj'} = ($FileName, $LineNumber, $SubscreenObj);
	
	$self->InterpretBlock($NestedBlockKey);
	
	! $job->{Died}
		|| return;
			
#			die 'CallSubScreen - @$self{qw(FileName LineNumber)  = '
#				.$job->dump( @$self{'FileName','LineNumber'} );
	
	@$self{ qw(FileName LineNumber SubscreenObj) }
		= @{ pop @{ $self->{SubscreenStack} } };

}
#----------------------------------------------------------------------
sub TextItem
{ 
	my ($self, $Text) 	= @_;
	my $job 						= $self->{job};
	
	my @Segments 				= split chr(4),$Text;
	my $html 						= '';
	
	foreach (@Segments)
	{
		if ($_ =~ /^(.){(.*?)}.$/s)
		{
			# An embedded expression.
			
			my ($QuoteChar, $Expr) = ($1, $2);
			my $Val = $self->ContextEval($Expr);
			! $job->{Died}
				|| return;
			next if $QuoteChar eq '{';
			
			$Val = HTML::Entities::encode_entities($Val)
				if $QuoteChar eq '"';
			$html .= $Val;
		}
		elsif ($_ =~ /<!--#(use|resume) (.+)-->/s)
		{
			# New source file.
			
				$self->{FileName} = $2;
		}
		else
			{ $html .= $_ };
	};
	
	$self->Emit($html);
	
}
#----------------------------------------------------------------------
sub Emit			# Literal text to emit.
{
	my ($self, $html) = @_;
	my $job						= $self->{job};
	
	$self->{OutputBuffer} .= $html; 
	return
		if $html =~ /^[ \n\t]+$/s;
	$html =~ s/\n/\n$TxtPrefix/sg;
	# $job->trace("$TxtPrefix$html")
}
#----------------------------------------------------------------------


#----------------------------------------------------------------------
sub Eval 	# This is the _MAIN context eval, overrriden when a subscreen is
					# being interpreted, or by a <my>.  
					# Call: 
					#				$SubscreenObj->Eval($job, $Expression).
					# The namespace include '$vars', '$job', $ParValTemp, and SubscreenObj's pars.
					# The pars are package globals in a sythesized one-object class. 
{
	my ($Self, $Expression) = @_;
	my $job = $Self->{job};
	my ($vars, $ParValTemp) = @$job{'vars','ParValTemp'};
	$job->{confirmeval} = 0;
	$Expression = "\$job->{confirmeval}++; $Expression";
eval($Expression) }
#----------------------------------------------------------------------
sub Name {""}		# This will also be overridden in any screenObj context.
#----------------------------------------------------------------------


#----------------------------------------------------------------------
sub ContextEval		# Evaluate string in name space defined by $SubscreenObj.
{
	my ( $self, $StringToEval, $SubscreenObj ) = @_;
	my $job 				= $self->{job};
	$SubscreenObj		||= $self->{SubscreenObj};
	
	my $result = $SubscreenObj->Eval($StringToEval);
	if ($@)
	{
		my $err = $@;
		eval(1);
		return $job->dieClean($err);
	};
	$job->{confirmeval}
		|| return $job->dieClean
			('Could not compile expression, usually due to sytax errors or undeclared variables:'
			."<br>\n$StringToEval");
	
	my $trace;
		if (! defined $result)
		{
			$trace = 'undef'
		}
		elsif (ref $result)
		{
			$trace = 'ref to '. ref $result
		}
		else
		{
			$trace = $result;
		};
		# $job->trace("ContextEval of '$StringToEval' = '$trace'");
	
$result}
#----------------------------------------------------------------------
sub StoreValueInContext
{
	my ($self, $lvalue, $value, $SubscreenObj) = @_;
	my $job 						= $self->{job};
	$job->{ContextVal} 	= $value;
	$self->ContextEval("$lvalue = \$job->{ContextVal}", $SubscreenObj);
	! $job->{Died}
		|| return $job->dieClean
			("Could not store '$value' into '$lvalue'");
1}
#----------------------------------------------------------------------
sub QuoteAttribute
{
	my ($self, $val) = @_;
	my $job = $self->{job};
	my $quote = $val =~ /"/s 
		?		( $val =~ /'/s 
						? return $job->dieClean
							("Attribute value ($val) contains both single and double quotes")
						: "'"
				)
		: '"';
"$quote$val$quote"}
#----------------------------------------------------------------------
sub NameListAttr	# Expect $Attr->{$AttrName} to be a list of identifiers
									# as a string, separated by / $,\t\n/.  Return an array of the identifiers
									# If '*' is specified, use the keys of $StarDefaultHash.
{
	my ($self, $AttrName, $IdPattern, $StarDefaultHash ) = @_;
	my $job 	= $self->{job};
	my $Attr 	= $self->{Attr};
	
	my $ws = '[ $\\n\\t,\\015]';
	
	my $namesString = $Attr->{$AttrName}
		|| return [];
		
	return  [keys %$StarDefaultHash]
		if ($StarDefaultHash && '*' eq $namesString);
		
	my @namesList = split "$ws+", $namesString;
	$namesList[0]
		|| shift @namesList;
	foreach (@namesList) 
		{ $_ =~ /$IdPattern/s || return $job->dieClean("'$_' not a valid identifier in '$namesString'") };
		
[@namesList] }
#----------------------------------------------------------------------
#
#	ezApp tag methods follow.
#
#	The parameters to these methods are provided in $self:
#		
#			FileName				=> source file of tag
#			LineNumber 			=> source line number of tag
#			Attr						=> hash, attr_name => attr_value
#			AttrSeq					=> sequence of attr_names as occurred in tag
#			NestedBlockKey	=> key of nested block, for block tags
#
#
#----------------------------------------------------------------------
sub gotoscreen		# Redirect to named screen.
{
	my $self = shift;
	my $Attr = $self->{Attr};
	my $job  = $self->{job};
	
	my $ScreenName = $Attr->{name}
		or $job->dieClean("'name' required");
		
	-e "ezscrs/$ScreenName.ezscr"
		or $job->dieClean("Could not find screen '$ScreenName'");
		
	$self->{RedirectFile} = "$ScreenName.ezscr";
	
}
#----------------------------------------------------------------------
sub ezbutton 	
{
	my $self = shift;
	my $Attr = $self->{Attr};
	
	# Since a button of a given name may occur multiple times, each generated instance
	# is assigned a $seqnum, beginning with 1.
	#
	# Emit:
	#			 	$query->submit
	#					( -name=>"$ObjectName:ezbutton:$Attr->{name}:$seqnum"
	#						-label=>$Attr->{label}
	#					);
	#
	# For attributes other than 'name' and 'label', save the value 
	# $Attr->{$key} in a hidden field named:
	#
	#				"$ObjectName:ezbutton:$Attr->{name}:$seqnum:$key".  
	#
	#	Include $TagText among these as $Attr->{tagtext}.
	#
	# This enables ezAPP::job (on the next invocation) to reconstruct $Attr
	# from $query->param() for whichever ezbutton is hit.
							
	my $job 				= $self->{job};
	my $vars				= $job->{vars};
	my $query 			= $job->{query};
	my $ObjectName 	= $self->{SubscreenObj}->Name;
	my $html;
	
	# Access and validate parameters.
	
		my $name 	= $Attr->{name}; 	delete $Attr->{name};
		my $label = $Attr->{label}; delete $Attr->{label};
		($name && $label)
			|| return $job->dieClean
				("Both 'name' and 'label' are required");
		$name !~ /:|\t|\n/s
			|| return $job->dieClean
				("ezbutton names may not contain [:\\t\\n]");
		my $QualifiedName = "$ObjectName:ezbutton:$name";


	# Emit our CGI button.
	
		my $seqnum = ++$job->{$QualifiedName};
		$html = $query->submit( -name=>"$QualifiedName:$seqnum", -value=>$label )."\n";
		
		
	# Verify screen existence if one specified.
		if ( exists $Attr->{screen} )
		{
			-e "ezscrs/$Attr->{screen}.ezscr"
				|| return $job->dieClean
					("Screen '$Attr->{screen}' could not be found");
		};
		
	# Emit hidden fields for attributes.
	
		( $Attr->{filename} || $Attr->{linenumber} )
			&& $job->dieClean("Button attributes 'filename' and 'linenumber' are reserved");
			
		$Attr->{filename} = $self->{FileName};
		$Attr->{linenumber} = $self->{LineNumber};
		foreach (keys %$Attr)
		{
			my $FieldName = "$QualifiedName:$seqnum:$_";
			$query->param( -name=>$FieldName, -value=>$Attr->{$_} );
			$html .= $query->hidden(-name=>$FieldName)."\n";
		};
	
	$self->Emit($html);
}
#----------------------------------------------------------------------
sub editvar
{
	my $self = shift;
	
	my ( $Attr, $AttrSeq ) = @$self{'Attr','AttrSeq'};
	my $job 							= $self->{job};
	my $query 						= $job->{query};
	my $vars 							= $job->{vars};
	my $ReservedVarNames 	= $job->{ReservedVarNames};
	
	# Validate parameters.
	
		my $FieldType = $Attr->{type} || 'textfield';
		delete $Attr->{type};
		my $var = $Attr->{name} 
			|| return $job->dieClean("'name' required");
		$var =~ /^$VarNamePattern$/s
			|| return $job->dieClean("'name' must match $VarNamePattern");
		( ! $ReservedVarNames->{$var} )
			|| return $job->dieClean("'$var' is a reserved var name");
		! exists $Attr->{value}
			|| return $job->dieClean
				("'value' not allowed - you can store into \$vars->{$var} or you can"
					."use 'choicearray', depending on your intent");
		my $choicearray = $Attr->{choicearray};
		delete $Attr->{choicearray};
		if ( $choicearray ) {
			'ARRAY' eq ref $choicearray
				|| return $job->dieClean
					("'choicearray' requires an array ref");
		};
	
	# Store current $vars value into $query->param.

			$query->param( -name=>$var, -value=>$vars->{$var} );
				
	my @QueryArgs = map { "-$_=>\"$Attr->{$_}\"" } keys %$Attr;
	( push @QueryArgs, '-value=>$choicearray' ) if $choicearray;
	my $ArgList = join ', ', @QueryArgs;
	my $QueryCall = "\$query->$FieldType( $ArgList )";
	my $html = eval $QueryCall 
		|| return $job->dieClean
			("Could not generate CGI code for editing: '$@'");
			
	$job->{editedvars}->{$var}++;
	$self->Emit($html);
}
#----------------------------------------------------------------------
sub persistent
{
	my ($self, $PersistentVal) 	= @_;
	my ($Attr, $AttrSeq) 				= @$self{'Attr','AttrSeq'};
	my $job 										= $self->{job};
	my $PersistentVars 					= $job->{PersistentVars};
	
	defined $PersistentVal
		|| ( $PersistentVal++ ) ;
		
	$Attr->{vars}
		|| return $job->dieClean("'vars' required");
		
	my $VarList = $self->NameListAttr('vars', $VarNamePattern, $job->{editedvars} );
	! $job->{Died}
	|| return;
	
	if ($PersistentVal)
		{ foreach (@$VarList) { $PersistentVars->{$_}++ } }
	else
		{ delete @$PersistentVars{@$VarList} };
}
#----------------------------------------------------------------------
sub nonpersistent
{
	my $self = shift;
	$self->persistent(@_, 0);
}
#----------------------------------------------------------------------
sub hiddenfields	
{
	my $self = shift;
	my $job = $self->{job};
	my $ezbuttons = $job->{ezbuttons};
	my ( $query, $vars, $PersistentVars, $EditedVars ) 
		= @$job{ qw( query vars PersistentVars editedvars ) };
	my $html = '';
	
	# Reformat persistency hashes for more efficient hidden storage.
	
		$vars->{persistentvars} = join ',', keys %$PersistentVars;
		$vars->{editedvars} 		= join ',', keys %$EditedVars;
		
	# Save those persistent vars which have not been edited.
		foreach my $var ( keys %$PersistentVars ) 
		{
			next if $EditedVars->{$var};
			$query->param( -name=>$var, -value=>$vars->{$var} );
			my $htmlHiddens = $query->hidden( -name=>$var );
			my @htmlHiddens = split '><', $htmlHiddens;
			$htmlHiddens = join ">\n<", @htmlHiddens;
			$html .= "$htmlHiddens\n";
		};
		
	$self->Emit($html);
}
#----------------------------------------------------------------------
#
#		Database methods.
#
#----------------------------------------------------------------------
sub dbopen		# <dbopen 
							#		dbhto				variable to hold db object ref
							#		name				name of database
							#	 	handler 		eg 'mysql'
							#		user				eg, 'root'
							#		password		
							#	>
{
	my $self 		= shift;
	my ($Attr, $AttrSeq) = @$self{'Attr','AttrSeq'};
	my $job 				= $self->{job};
	
	my ($dbhto, $name, $handler, $user, $password) =
		map {$_ ? $_ : ''}
			@$Attr{ qw(dbhto name handler user password) };
	($dbhto && $name && $user) 
		|| return $job->dieClean
			("'dbhto name user' are all required");
	
#	return $job->dieClean("dbopen: '$handler','$name','$user','$password'");
			
	my $dbh = ezApp::DB->new($job, $name, $user, $handler, $password);
	my $result = $dbh->{isopen} ? $dbh : '';
	$self->StoreValueInContext($dbhto, $result)
}
#----------------------------------------------------------------------
sub dbclose
{	
	my $self = shift;
	my $dbh = $self->Get_dbh
		|| return;
	$dbh->Close;
}
#----------------------------------------------------------------------
sub Get_dbh		# This is not a tag method.
{
	my $self 							= shift;
	my $job 							= $self->{job};
	my ($Attr, $AttrSeq) 	= @$self{'Attr','AttrSeq'};
	my $dbh 							= $Attr->{dbh}
		|| return $job->dieClean("'dbh' required");
	'ezApp::DB' eq ref $dbh
		|| return $job->dieClean
			("'dbh' is not a DB object reference");
$dbh}
#----------------------------------------------------------------------
sub dbgetrecords		# <dbgetrecords 
										#		dbh 
										#		recordsto
										#		table 
										#		columns
										#		where					optional
										#		orderby				optional
										#		limit					optional
										#	>
{
	my $self 					= shift;
	my ($Attr, $AttrSeq) = @$self{'Attr','AttrSeq'};
	my $job 					= $self->{job};
	
	my $dbh = $self->Get_dbh
		|| return;
	
	my ($recordsto, $table, $columns, $where, $orderby, $limit) =
		map {$_ ? $_ : ''}
			@$Attr{ qw(recordsto table columns where orderby limit) };
	($table && $columns && $recordsto)
		|| return $job->dieClean 
			("'table columns recordsto' are all required");
	'ARRAY' eq ref $columns
		|| return  $job->dieClean
			("'columns' is not an array ref, perhaps you meant 'columns_'");
	
	my $ParamHash ={};
	$ParamHash->{Where} = $where if $where;		
	$ParamHash->{OrderBy} = $orderby if $orderby;		
	$ParamHash->{Limit} = $limit if $limit;		
			
	my $records = $dbh->Records( $table, -Columns=>$columns, $job->HashAsArgs($ParamHash) );
	$self->StoreValueInContext($recordsto, $records)
}
#----------------------------------------------------------------------
sub dbupdate		# <dbupdate 
								#		dbh 
								#		table 
								#		columns
								#		values
								#		where
								#	>
{	
	my $self 					= shift;
	my $job 					= $self->{job};
	my $dbh 					= $self->Get_dbh
		|| return;
	my ($table, $columns, $values, $where) = map {$_ ? $_ : ''} 
		@{ $self->{Attr} }{qw(table columns values where)};
	$table && $columns && $values && $where
		|| return $job->dieClean
			("'table columns values where' all required");
	$dbh->UpdateRecords($table, -Columns=>$columns, -Values=>$values, -Where=>$where)
}
#----------------------------------------------------------------------
sub dbdo				# <dbdo
								#		dbh 
								#		statement 
								#		resultto		optional
								#	>
{	
	my $self 							= shift;
	my ($Attr, $AttrSeq) 	= @$self{'Attr','AttrSeq'};
	my $job 							= $self->{job};
	my $dbh 							= $self->Get_dbh
		|| return;
	exists $Attr->{statement}
		|| return $job->dieClean("'statement' required");
	my $result = $dbh->DoSQL($Attr->{statement});
	return unless exists $Attr->{resultto};
	$self->StoreValueInContext
		($Attr->{resultto}, $result)
}
#----------------------------------------------------------------------
sub dblastinsertid		# <dblastinsertid
											#		dbh 
											#		idto 
											#	>

{	
	my $self 							= shift;
	my ($Attr, $AttrSeq) 	= @$self{'Attr','AttrSeq'};
	my $job 							= $self->{job};
	my $dbh 							= $self->Get_dbh(@_)
		|| return;
		
	exists $Attr->{idto}
		|| return $job->dieClean("'idto' required");
		
	$self->StoreValueInContext
		($Attr->{idto}, $dbh->LastInsertId)
}
#----------------------------------------------------------------------
sub TagOutOfContext # (not a tag method.)
{
	my $self 	= shift;
	my $job 	= $self->{job};
	return $job->dieClean
		("<$self->{Tag}> not valid");
''} 
#----------------------------------------------------------------------
#
#		Methods that deal with conditionals
#
#----------------------------------------------------------------------
sub Condition		# Process a 'true' or 'false' attribute. 
{
	my ($self, $Attr) = @_;
	my $job 					= $self->{job};
	$Attr 						||= $self->{Attr};
	
	my $Condition;
	if ( exists $Attr->{true} )
	{
		$Condition = $Attr->{true};		
	} 
	elsif ( exists $Attr->{false} )
	{
		$Condition = ! $Attr->{false};
	} 
	else
	{
		return $job->dieClean("'true' or 'false' required");
	};
	
$Condition}
#----------------------------------------------------------------------
sub if
{
	my $self 							= shift;
	my $job								= $self->{job};
	my $NestedBlockKey		= $self->{NestedBlockKey};
	
	$self->{IfCaseDone} 	= '';
	
	my $Condition 				= $self->Condition ;
	! $job->{Died}
		|| return;
		
	return unless $Condition;
	
	$self->InterpretBlock($NestedBlockKey);
	$self->{IfCaseDone}++;
	
}
#----------------------------------------------------------------------
sub else
{
	my $self = shift;
	
	return if $self->{IfCaseDone};
	
	my $NestedBlockKey = $self->{NestedBlockKey};
	$self->InterpretBlock($NestedBlockKey);
	$self->{IfCaseDone}++;
}
#----------------------------------------------------------------------
sub elsif
{
	my $self 							= shift;
	my $job								= $self->{job};
		
	return if $self->{IfCaseDone};
	
	my $Condition 				= $self->Condition;
	! $job->{Died}
		|| return;
		
	return unless $Condition;
	
	my $NestedBlockKey = $self->{NestedBlockKey};
	$self->InterpretBlock($NestedBlockKey);
	$self->{IfCaseDone}++;
}
#----------------------------------------------------------------------
#
#		Methods dealing with loops
#
#----------------------------------------------------------------------
sub repeat
{
	my $self 	= shift;
	my $job 	= $self->{job};
	my ( $Tag, $Attr, $NestedBlockKey ) = @$self->{'Tag','Attr','NestedBlockKey'};
	$self->{LoopDepth}++;
	$self->{Terminator} = '';
	
	my $Times;
	( exists $Attr->{times} && ! ref ($Times = $Attr->{times}) )
		|| return $job->dieClean("integer 'times' required");
	$Times > 0
		|| ( $Times = 0 );
	
	foreach (1..$Times)
	{
		$self->InterpretBlock($NestedBlockKey);
		return if $job->{Died};
		my $Terminator = $self->{Terminator};
		$self->{Terminator} = '';
		next if $Terminator eq 'next';
		last if $Terminator eq 'last';
	};
	
	$self->{LoopDepth}--;
}
#----------------------------------------------------------------------
sub EvalAttr
{
	my ($self, $Attr) = @_;
	my $job						= $self->{job};
	
	foreach (keys %$Attr)
	{
		my $AttrVal = $Attr->{$_};
		
		if ($AttrVal =~ /^{(.+?)}$/s)
		{
			# AttrVal is an expression.
			
				$AttrVal = $self->ContextEval($1);
				! $job->{Died}
					|| return;
			
			$Attr->{$_} = $AttrVal;
		};
	};
}
#----------------------------------------------------------------------
sub while
{
	my $self 	= shift;
	my $job		= $self->{job};
	my ($RawAttr, $NestedBlockKey) 
		= @$self{'Attr','NestedBlockKey'};
	$self->{LoopDepth}++;
	$self->{Terminator} = '';
	
	while(1)
	{
		my $Attr = {%$RawAttr};
		$self->EvalAttr($Attr);
		! $job->{Died}
			|| return;
		last unless $self->Condition($Attr);
		$self->InterpretBlock($NestedBlockKey);
		return if $job->{Died};
		my $Terminator = $self->{Terminator};
		$self->{Terminator} = '';
		next if $Terminator eq 'next';
		last if $Terminator eq 'last';
	};
	
	$self->{LoopDepth}--;
}
#----------------------------------------------------------------------
sub foreach
{
	my $self 				= shift;
	my $job					= $self->{job};
	my ($Attr, $NestedBlockKey) = @$self{'Attr','NestedBlockKey'};
	$self->{LoopDepth}++;
	$self->{Terminator} = '';
		
	exists $Attr->{item} && exists $Attr->{fromarray}
		|| return $job->dieClean("'item' and 'fromarray' required");
		
	my $Item = $Attr->{item};
	delete $Attr->{item};

	$self->EvalAttr($Attr);
		
	'ARRAY' eq ref $Attr->{fromarray}
		|| return $job->dieClean
			("'fromarray' must give an array ref");
	
	foreach ( @{ $Attr->{fromarray} } )
	{
		$job->{xx} = $_;
		$self->ContextEval("$Item = \$job->{xx}");
		! $job->{Died}
			|| return;
		$self->InterpretBlock($NestedBlockKey);
		return if $job->{Died};
		my $Terminator = $self->{Terminator};
		$self->{Terminator} = '';
		next if $Terminator eq 'next';
		last if $Terminator eq 'last';
	};
	
	$self->{LoopDepth}--;
}
#----------------------------------------------------------------------
sub next
{
	my $self 	= shift;
	my $job 	= $self->{job};
	$self->{LoopDepth}
		|| return $job->dieClean("<next> must be within a loop");
	
	$self->{Terminator} = 'next';
}
#----------------------------------------------------------------------
sub last
{
	my $self 	= shift;
	my $job 	= $self->{job};
	$self->{LoopDepth}
		|| return $job->dieClean("<last> must be within a loop");
	
	$self->{Terminator} = 'last';
}
#----------------------------------------------------------------------
#
#		Methods dealing with Subscreens
#
#----------------------------------------------------------------------
#
#		Temlate for creating dynamic packages which provide a namespace
#		for each subscreen call... providing recursive tags.
#
my $PackageTemplate = <<'---'
package $PackageName;
$my_of_ParNames
sub new
{
	my ($class, $job) = @_;
	my $self = {job=>$job, Name=>$PackageName};
	bless $self, $class;
}
#----------------------------------------------------------------------
sub Eval
{ 
	my ($Self, $Expression) = @_;
	my $job = $Self->{job};
	my ($vars, $ParValTemp) = @$job{'vars','ParValTemp'};
	$job->{confirmeval} = 0;
	$Expression = "\$job->{confirmeval}++; $Expression";
eval($Expression) }
#----------------------------------------------------------------------
sub Name { shift()->{Name} }
#----------------------------------------------------------------------
1;
---
;
#----------------------------------------------------------------------
sub screen
{
	my $self = shift;
	my ($Attr, $NestedBlockKey) = @$self{'Attr','NestedBlockKey'};
	
	my $job 							= $self->{job};
	my $vars							= $job->{vars};
	my $ValuesAttr 				= {};
	
	my $ParNames = $self->NameListAttr('pars', $ParNamePattern);
	! $job->{Died}
		|| return;
	
	# Seek (possibly undefined) values for pars in $vars.
		
		foreach (@$ParNames)
		{
			$ValuesAttr->{$_} = $vars->{"ezpar:$_"}
		};
	
	$self->{SubscreenObj} = $self->SetupParSpace($ParNames, '_MAIN', $ValuesAttr);
	
	$self->InterpretBlock($NestedBlockKey);
}
#----------------------------------------------------------------------
sub subscreen
{
	my $self 													= shift;
	my ($Attr, $NestedBlockKey) 			= @$self{'Attr','NestedBlockKey'};
	my $job 													= $self->{job};
	my $Subscreens 										= $self->{Subscreens};
		
	$Attr->{name}
			|| return $job->dieClean("'name' required");
	$Attr->{name} =~ /^[a-z^][a-z0-9^]*$/s
		|| return $job->dieClean("'name' must match /[a-z][a-z0-9]*/s");
	my $SubscreenName = $Attr->{name};
	
	# Make sure subscreen name not already registered.
	
		! $Subscreens->{$SubscreenName}
			|| return $job->dieClean("Duplicate subscreen name");
		
	my $ParNames = $self->NameListAttr('pars', $ParNamePattern);
	! $job->{Died}
		|| return;
		
	@{ $Subscreens->{$SubscreenName} }
		{ qw(BodyKey Pars FileName LineNumber) }
			= ($NestedBlockKey, $ParNames, $self->{FileName}, $self->{LineNumber} );


#	$Subscreens->{$SubscreenName} = [ $NestedBlockKey, $ParNames, $self->{FileName}, $self->{LineNumber} ];
		
}
#----------------------------------------------------------------------
sub SetupParSpace			# Instantiate a new package for a screen or subscreen.
											# $ParNames gives the names of the pars, and $ValuesAttr 
											# provides whatever values have been set for them.
{
	my ($self, $ParNames, $SubscreenName, $ValuesAttr) = @_;
	my $job 				= $self->{job};
	my $Subscreens 	= $self->{Subscreens};
	
	# Instantiate a package for our subscreen.
	
		my $my_of_ParNames = join ',', map{"\$$_"} @$ParNames;
		$my_of_ParNames = $my_of_ParNames
			? "my ($my_of_ParNames);"
				: '';
		my $PackageName = $SubscreenName.( ++$Subscreens->{$SubscreenName}->{CallSeqNum} );
		my $PackageCreator = $PackageTemplate;
		$PackageCreator =~ s/\$PackageName/$PackageName/s;
		$PackageCreator =~ s/\$my_of_ParNames/$my_of_ParNames/s;
		eval "$PackageCreator"
			|| return $job->dieClean
				("Could not create package for '$SubscreenName'");
		my $SubscreenObj = $PackageName->new($job)
			|| return $job->dieClean
				("Could not create object for '$SubscreenName'");
		
	# Assign calling Attrs to subscreen pars.
	
		foreach (@$ParNames)
		{
			$self->StoreValueInContext("\$$_", $ValuesAttr->{$_}, $SubscreenObj);
		};
	
		
$SubscreenObj }
#----------------------------------------------------------------------
1;
